from .sdk.client import Laminar
from .sdk.decorators import observe, lmnr_context, wrap_llm_call
from .sdk.interface import trace, TraceContext, SpanContext
from .sdk.tracing_types import EvaluateEvent
from .sdk.types import ChatMessage, PipelineRunError, PipelineRunResponse, NodeInput

from .semantic_conventions import *



from typing import Any, Literal, Optional, Union
import datetime
import pydantic
import uuid

from .constants import CURRENT_TRACING_VERSION
from .utils import to_dict


class EvaluateEvent(pydantic.BaseModel):
    name: str
    data: str
    timestamp: Optional[datetime.datetime] = None


class Span(pydantic.BaseModel):
    version: str = CURRENT_TRACING_VERSION
    spanType: Literal["DEFAULT", "LLM"] = "DEFAULT"
    id: uuid.UUID
    traceId: uuid.UUID
    parentSpanId: Optional[uuid.UUID] = None
    name: str
    # generated at start of span, so required
    startTime: datetime.datetime
    # generated at end of span, optional when span is still running
    endTime: Optional[datetime.datetime] = None
    attributes: dict[str, Any] = {}
    input: Optional[Any] = None
    output: Optional[Any] = None
    metadata: Optional[dict[str, Any]] = None
    evaluateEvents: list[EvaluateEvent] = []
    events: list["Event"] = None

    def __init__(
        self,
        name: str,
        trace_id: uuid.UUID,
        start_time: Optional[datetime.datetime] = None,
        version: str = CURRENT_TRACING_VERSION,
        span_type: Literal["DEFAULT", "LLM"] = "DEFAULT",
        id: Optional[uuid.UUID] = None,
        parent_span_id: Optional[uuid.UUID] = None,
        input: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = {},
        attributes: Optional[dict[str, Any]] = {},
        evaluate_events: list[EvaluateEvent] = [],
    ):
        super().__init__(
            version=version,
            spanType=span_type,
            id=id or uuid.uuid4(),
            traceId=trace_id,
            parentSpanId=parent_span_id,
            name=name,
            startTime=start_time or datetime.datetime.now(datetime.timezone.utc),
            input=input,
            metadata=metadata or {},
            attributes=attributes or {},
            evaluateEvents=evaluate_events,
            events=[],
        )

    def update(
        self,
        end_time: Optional[datetime.datetime],
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        evaluate_events: Optional[list[EvaluateEvent]] = None,
        override: bool = False,
    ):
        self.endTime = end_time or datetime.datetime.now(datetime.timezone.utc)
        self.input = input
        self.output = output
        new_metadata = (
            metadata if override else {**(self.metadata or {}), **(metadata or {})}
        )
        new_attributes = (
            attributes or {}
            if override
            else {**(self.attributes or {}), **(attributes or {})}
        )
        new_evaluate_events = (
            evaluate_events or []
            if override
            else self.evaluateEvents + (evaluate_events or [])
        )
        self.metadata = new_metadata
        self.attributes = new_attributes
        self.evaluateEvents = new_evaluate_events

    def add_event(self, event: "Event"):
        self.events.append(event)

    def to_dict(self) -> dict[str, Any]:
        try:
            obj = self.model_dump()
        except TypeError:
            # if inner values are pydantic models, we need to call model_dump on them
            # see: https://github.com/pydantic/pydantic/issues/7713
            obj = {}
            for key, value in self.__dict__.items():
                obj[key] = (
                    value.model_dump()
                    if isinstance(value, pydantic.BaseModel)
                    else value
                )
        obj = to_dict(obj)
        return obj


class Trace(pydantic.BaseModel):
    id: uuid.UUID
    version: str = CURRENT_TRACING_VERSION
    success: bool = True
    userId: Optional[str] = None  # provided by user or null
    sessionId: Optional[str] = None  # provided by user or uuid()
    release: Optional[str] = None
    metadata: Optional[dict[str, Any]] = None

    def __init__(
        self,
        success: bool = True,
        id: Optional[uuid.UUID] = None,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ):
        id_ = id or uuid.uuid4()
        super().__init__(
            id=id_,
            success=success,
            userId=user_id,
            sessionId=session_id,
            release=release,
            metadata=metadata or {},
        )

    def to_dict(self) -> dict[str, Any]:
        try:
            obj = self.model_dump()
        except TypeError:
            # if inner values are pydantic models, we need to call model_dump on them
            # see: https://github.com/pydantic/pydantic/issues/7713
            obj = {}
            for key, value in self.__dict__.items():
                obj[key] = (
                    value.model_dump()
                    if isinstance(value, pydantic.BaseModel)
                    else value
                )
        obj = to_dict(obj)
        return obj


class Event(pydantic.BaseModel):
    id: uuid.UUID
    templateName: str
    timestamp: datetime.datetime
    spanId: uuid.UUID
    value: Optional[Union[int, str, float, bool]] = None

    def __init__(
        self,
        name: str,
        span_id: uuid.UUID,
        timestamp: Optional[datetime.datetime] = None,
        value: Optional[Union[int, str, float, bool]] = None,
    ):
        super().__init__(
            id=uuid.uuid4(),
            templateName=name,
            spanId=span_id,
            timestamp=timestamp or datetime.datetime.now(datetime.timezone.utc),
            value=value,
        )

    def to_dict(self) -> dict[str, Any]:
        try:
            obj = self.model_dump()
        except TypeError:
            # if inner values are pydantic models, we need to call model_dump on them
            # see: https://github.com/pydantic/pydantic/issues/7713
            obj = {}
            for key, value in self.__dict__.items():
                obj[key] = (
                    value.model_dump()
                    if isinstance(value, pydantic.BaseModel)
                    else value
                )
        obj = to_dict(obj)
        return obj



from .base import Provider
from ...semantic_conventions.gen_ai_spans import (
    FINISH_REASONS,
    FREQUENCY_PENALTY,
    INPUT_TOKEN_COUNT,
    MAX_TOKENS,
    OUTPUT_TOKEN_COUNT,
    PRESENCE_PENALTY,
    REQUEST_MODEL,
    RESPONSE_MODEL,
    STOP_SEQUENCES,
    STREAM,
    TEMPERATURE,
    TOP_P,
    TOTAL_TOKEN_COUNT,
)
from .utils import parse_or_dump_to_dict

from collections import defaultdict
from typing import Any, Optional, Union
import logging
import pydantic

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk


class OpenAI(Provider):
    logger = logging.getLogger("lmnr.sdk.tracing.providers.openai")

    def display_name(self) -> str:
        return "OpenAI"

    def stream_list_to_dict(
        self, response: list[Union[ChatCompletionChunk, str]]
    ) -> dict[str, Any]:
        model = None
        finish_reasons = []
        output_tokens = 0
        outputs = defaultdict(lambda: defaultdict(str))
        try:
            for chunk in response:
                chunk = parse_or_dump_to_dict(chunk)
                if model is None:
                    model = chunk["model"]
                finish_reasons = [
                    choice.get("finish_reason") for choice in chunk.get("choices", [])
                ]
                for i, choice in enumerate(chunk.get("choices", [])):
                    if choice["delta"] and isinstance(choice["delta"], dict):
                        for key in choice["delta"]:
                            if choice["delta"][key] is None:
                                if key not in outputs[i]:
                                    outputs[i][key] = None
                                continue
                            outputs[i][key] += choice["delta"][key]
                        output_tokens += 1
        except Exception as e:
            self.logger.error(f"Error parsing streamming response: {e}")
            pass

        output_key_values = [
            self._message_to_key_and_output(dict(outputs[i]))
            for i in range(len(outputs))
        ]
        return {
            "model": model,
            "prompt_tokens": 0,
            "usage": {
                "prompt_tokens": 0,
                "completion_tokens": output_tokens,
                "total_tokens": output_tokens,
                "finish_reason": self._from_singleton_list(finish_reasons),
            },
            "choices": [
                (
                    {
                        "message": {
                            output[0]: output[1],
                            "role": "assistant",
                        }
                    }
                    if output
                    else None
                )
                for output in output_key_values
            ],
        }

    def extract_llm_attributes_from_response(
        self, response: Union[str, dict[str, Any], pydantic.BaseModel]
    ) -> dict[str, Any]:
        obj = parse_or_dump_to_dict(response)

        choices = obj.get("choices", [])
        decisions = []
        for choice in choices:
            # choice = parse_or_dump_to_dict(choice)
            if choice.get("content"):
                decisions.append("completion")
            elif choice.get("refusal"):
                decisions.append("refusal")
            elif choice.get("tool_calls"):
                decisions.append("tool_calls")
            else:
                decisions.append(None)

        return {
            RESPONSE_MODEL: obj.get("model"),
            INPUT_TOKEN_COUNT: obj.get("usage", {}).get("prompt_tokens"),
            OUTPUT_TOKEN_COUNT: obj.get("usage", {}).get("completion_tokens"),
            TOTAL_TOKEN_COUNT: obj.get("usage", {}).get("total_tokens"),
            FINISH_REASONS: obj.get("finish_reason"),
            # "decision": self._from_singleton_list(decisions),
        }

    def extract_llm_output(
        self, result: Union[str, dict[str, Any], ChatCompletion]
    ) -> Any:
        result = parse_or_dump_to_dict(result)
        choices = result.get("choices")
        if not choices:
            return None
        outputs = [choice["message"] for choice in choices]

        return self._from_singleton_list(outputs)

    def extract_llm_attributes_from_args(
        self, func_args: list[Any], func_kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        return {
            REQUEST_MODEL: func_kwargs.get("model"),
            TEMPERATURE: func_kwargs.get("temperature"),
            TOP_P: func_kwargs.get("top_p"),
            FREQUENCY_PENALTY: func_kwargs.get("frequency_penalty"),
            PRESENCE_PENALTY: func_kwargs.get("presence_penalty"),
            STOP_SEQUENCES: func_kwargs.get("stop"),
            MAX_TOKENS: func_kwargs.get("max_tokens"),
            STREAM: func_kwargs.get("stream", False),
        }

    def _message_to_key_and_output(
        self, message: Union[dict[str, Any], ChatCompletionMessage]
    ) -> Optional[tuple[str, str]]:
        message = parse_or_dump_to_dict(message)

        for key in ["content", "refusal", "tool_calls"]:
            if message.get(key) is not None:
                return (key, message[key])
        return None

    def _from_singleton_list(self, obj: Any) -> Any:
        # OpenAI returns list of choices. This will have more than one item
        # only if request parameter `n` is specified and is grate than 1.
        # That's a rare case, so we return the [contents of the] choice alone if there is just one.
        if isinstance(obj, list) and len(obj) == 1:
            return obj[0]
        return obj



import requests
import pydantic
import uuid
from typing import Optional, Union

from .utils import to_dict


class ChatMessage(pydantic.BaseModel):
    role: str
    content: str


class ConditionedValue(pydantic.BaseModel):
    condition: str
    value: "NodeInput"


NodeInput = Union[str, list[ChatMessage], ConditionedValue]  # TypeAlias


class PipelineRunRequest(pydantic.BaseModel):
    inputs: dict[str, NodeInput]
    pipeline: str
    env: dict[str, str] = pydantic.Field(default_factory=dict)
    metadata: dict[str, str] = pydantic.Field(default_factory=dict)
    stream: bool = pydantic.Field(default=False)
    parent_span_id: Optional[uuid.UUID] = pydantic.Field(default=None)
    trace_id: Optional[uuid.UUID] = pydantic.Field(default=None)

    # uuid is not serializable by default, so we need to convert it to a string
    def to_dict(self):
        return {
            "inputs": {
                k: v.model_dump() if isinstance(v, pydantic.BaseModel) else to_dict(v)
                for k, v in self.inputs.items()
            },
            "pipeline": self.pipeline,
            "env": self.env,
            "metadata": self.metadata,
            "stream": self.stream,
            "parentSpanId": str(self.parent_span_id) if self.parent_span_id else None,
            "traceId": str(self.trace_id) if self.trace_id else None,
        }


class PipelineRunResponse(pydantic.BaseModel):
    outputs: dict[str, dict[str, NodeInput]]
    run_id: str


class PipelineRunError(Exception):
    error_code: str
    error_message: str

    def __init__(self, response: requests.Response):
        try:
            resp_json = response.json()
            self.error_code = resp_json["error_code"]
            self.error_message = resp_json["error_message"]
            super().__init__(self.error_message)
        except Exception:
            super().__init__(response.text)

    def __str__(self) -> str:
        try:
            return str(
                {"error_code": self.error_code, "error_message": self.error_message}
            )
        except Exception:
            return super().__str__()



import abc
import pydantic
import typing


class Provider(abc.ABC):
    def display_name(self) -> str:
        raise NotImplementedError("display_name not implemented")

    def stream_list_to_dict(self, response: list[typing.Any]) -> dict[str, typing.Any]:
        raise NotImplementedError("stream_list_to_dict not implemented")

    def extract_llm_attributes_from_response(
        self, response: typing.Union[str, dict[str, typing.Any], pydantic.BaseModel]
    ) -> dict[str, typing.Any]:
        raise NotImplementedError(
            "extract_llm_attributes_from_response not implemented"
        )

    def extract_llm_output(
        self, response: typing.Union[str, dict[str, typing.Any], pydantic.BaseModel]
    ) -> typing.Any:
        raise NotImplementedError("extract_llm_output not implemented")

    def extract_llm_attributes_from_args(
        self, func_args: list[typing.Any], func_kwargs: dict[str, typing.Any]
    ) -> dict[str, typing.Any]:
        raise NotImplementedError("_extract_llm_attributes_from_args not implemented")



from .context import LaminarSingleton
from .tracing_types import EvaluateEvent, Span, Trace, Event

from typing import Any, Literal, Optional, Union
import datetime
import logging
import uuid


laminar = LaminarSingleton().get()


class ObservationContext:
    observation: Union[Span, Trace] = None
    _parent: "ObservationContext" = None
    _children: dict[uuid.UUID, "ObservationContext"] = {}
    _log = logging.getLogger("laminar.observation_context")

    def __init__(self, observation: Union[Span, Trace], parent: "ObservationContext"):
        self.observation = observation
        self._parent = parent
        self._children = {}

    def _get_parent(self) -> "ObservationContext":
        raise NotImplementedError

    def update(self, *args, **kwargs):
        raise NotImplementedError

    def span(
        self,
        name: str,
        input: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        span_type: Literal["DEFAULT", "LLM"] = "DEFAULT",
    ) -> "SpanContext":
        """Create a span within the current (trace or span) context.

        Args:
            name (str): Span name
            input (Optional[Any], optional): Inputs to the span. Defaults to None.
            metadata (Optional[dict[str, Any]], optional): Any additional metadata. Defaults to None.
            attributes (Optional[dict[str, Any]], optional): Any pre-defined attributes. Must comply to the convention. Defaults to None.
            span_type (Literal[&quot;DEFAULT&quot;, &quot;LLM&quot;], optional): Type of the span. Defaults to "DEFAULT".

        Returns:
            SpanContext: The new span context
        """
        parent = self
        parent_span_id = (
            parent.observation.id if isinstance(parent.observation, Span) else None
        )
        trace_id = (
            parent.observation.traceId
            if isinstance(parent.observation, Span)
            else parent.observation.id
        )
        span = laminar.create_span(
            name=name,
            trace_id=trace_id,
            input=input,
            metadata=metadata,
            attributes=attributes,
            parent_span_id=parent_span_id,
            span_type=span_type,
        )
        span_context = SpanContext(span, self)
        self._children[span.id] = span_context
        return span_context

    def id(self) -> uuid.UUID:
        """Get the uuid of the current observation

        Returns:
            uuid.UUID: UUID of the observation
        """
        return self.observation.id


class SpanContext(ObservationContext):
    def _get_parent(self) -> ObservationContext:
        return self._parent

    def end(
        self,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        evaluate_events: Optional[list[EvaluateEvent]] = None,
        override: bool = False,
    ) -> "SpanContext":
        """End the span with the given output and optional metadata and evaluate events.

        Args:
            input (Optional[Any], optional): Inputs to the span. Defaults to None.
            output (Optional[Any], optional): output of the span. Defaults to None.
            metadata (Optional[dict[str, Any]], optional): any additional metadata to the span. Defaults to None.
            attributes (Optional[dict[str, Any]], optional): pre-defined attributes (see semantic-convention). Defaults to None.
            check_event_names (Optional[list[EvaluateEvent]], optional): List of events to evaluate for and tag. Defaults to None.
            override (bool, optional): override existing metadata fully. If False, metadata is merged. Defaults to False.

        Returns:
            SpanContext: the finished span context
        """
        if self._children:
            self._log.warning(
                "Ending span %s, but it has children that have not been finalized. Children: %s",
                self.observation.name,
                [child.observation.name for child in self._children.values()],
            )
        self._get_parent()._children.pop(self.observation.id)
        return self._update(
            input=input,
            output=output,
            metadata=metadata,
            evaluate_events=evaluate_events,
            attributes=attributes,
            override=override,
            finalize=True,
        )

    def update(
        self,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        evaluate_events: Optional[list[EvaluateEvent]] = None,
        override: bool = False,
    ) -> "SpanContext":
        """Update the current span with (optionally) the given output and optional metadata and evaluate events, but don't end it.

        Args:
            input (Optional[Any], optional): Inputs to the span. Defaults to None.
            output (Optional[Any], optional): output of the span. Defaults to None.
            metadata (Optional[dict[str, Any]], optional): any additional metadata to the span. Defaults to None.
            attributes (Optional[dict[str, Any]], optional): pre-defined attributes (see semantic-convention). Defaults to None.
            check_event_names (Optional[list[EvaluateEvent]], optional): List of events to evaluate for and tag. Defaults to None.
            override (bool, optional): override existing metadata fully. If False, metadata is merged. Defaults to False.

        Returns:
            SpanContext: the finished span context
        """
        return self._update(
            input=input or self.observation.input,
            output=output or self.observation.output,
            metadata=metadata,
            evaluate_events=evaluate_events,
            attributes=attributes,
            override=override,
            finalize=False,
        )

    def event(
        self,
        name: str,
        value: Optional[Union[str, int, float, bool]] = None,
        timestamp: Optional[datetime.datetime] = None,
    ) -> "SpanContext":
        """Associate an event with the current span

        Args:
            name (str): name of the event. Must be predefined in the Laminar events page.
            value (Optional[Union[str, int, float, bool]], optional): value of the event. Must match range definition in Laminar events page. Defaults to None.
            timestamp (Optional[datetime.datetime], optional): If you need custom timestamp. If not specified, current time is used. Defaults to None.

        Returns:
            SpanContext: the updated span context
        """
        event = Event(
            name=name,
            span_id=self.observation.id,
            timestamp=timestamp,
            value=value,
        )
        self.observation.add_event(event)
        return self

    def evaluate_event(self, name: str, data: str) -> "SpanContext":
        """Evaluate an event with the given name and data. The event value will be assessed by the Laminar evaluation engine.
        Data is passed as an input to the agent, so you need to specify which data you want to evaluate. Most of the times,
        this is an output of the LLM generation, but sometimes, you may want to evaluate the input or both. In the latter case,
        concatenate the input and output annotating with natural language.

        Args:
            name (str): Name of the event. Must be predefined in the Laminar events page.
            data (str): Data to be evaluated. Typically the output of the LLM generation.

        Returns:
            SpanContext: the updated span context
        """
        self._update(
            input=self.observation.input,
            output=self.observation.output,
            evaluate_events=[
                EvaluateEvent(
                    name=name,
                    data=data,
                    timestamp=datetime.datetime.now(datetime.timezone.utc),
                )
            ],
            override=False,
        )

    def _update(
        self,
        input: Optional[Any] = None,
        output: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        evaluate_events: Optional[list[EvaluateEvent]] = None,
        override: bool = False,
        finalize: bool = False,
    ) -> "SpanContext":
        self.observation = laminar.update_span(
            input=input,
            output=output,
            span=self.observation,
            end_time=datetime.datetime.now(datetime.timezone.utc),
            metadata=metadata,
            attributes=attributes,
            evaluate_events=evaluate_events,
            finalize=finalize,
            override=override,
        )
        return self


class TraceContext(ObservationContext):
    def _get_parent(self) -> "ObservationContext":
        return self

    def update(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        success: bool = True,
    ) -> "TraceContext":
        """Update the current trace with the given metadata and success status.

        Args:
            user_id (Optional[str], optional): Custom user_id of your user. Useful for grouping and further analytics. Defaults to None.
            session_id (Optional[str], optional): Custom session_id for your session. Random UUID is generated on Laminar side, if not specified.
                                                  Defaults to None.
            release (Optional[str], optional): _description_. Release of your application. Useful for grouping and further analytics. Defaults to None.
            metadata (Optional[dict[str, Any]], optional):  any additional metadata to the trace. Defaults to None.
            success (bool, optional): whether this trace ran successfully. Defaults to True.

        Returns:
            TraceContext: updated trace context
        """
        return self._update(
            user_id=user_id or self.observation.userId,
            session_id=session_id or self.observation.sessionId,
            release=release or self.observation.release,
            metadata=metadata or self.observation.metadata,
            success=success if success is not None else self.observation.success,
        )

    def _update(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        success: bool = True,
        end_time: Optional[datetime.datetime] = None,
    ) -> "TraceContext":
        self.observation = laminar.update_trace(
            id=self.observation.id,
            user_id=user_id,
            session_id=session_id,
            release=release,
            metadata=metadata,
            success=success,
        )
        return self


def trace(
    user_id: Optional[str] = None,
    session_id: Optional[str] = None,
    release: Optional[str] = None,
) -> TraceContext:
    """Create the initial trace context. All further spans will be created within this context.

    Args:
        user_id (Optional[str], optional): Custom user_id of your user. Useful for grouping and further analytics. Defaults to None.
        session_id (Optional[str], optional): Custom session_id for your session. Random UUID is generated on Laminar side, if not specified.
                                                Defaults to None.
        release (Optional[str], optional): _description_. Release of your application. Useful for grouping and further analytics. Defaults to None.

    Returns:
        TraceContext: the pointer to the trace context. Use `.span()` to create a new span within this context.
    """
    session_id = session_id or str(uuid.uuid4())
    trace_id = uuid.uuid4()
    trace = laminar.update_trace(
        id=trace_id,
        user_id=user_id,
        session_id=session_id,
        release=release,
    )
    return TraceContext(trace, None)



from .tracing_types import Span, Trace

from pydantic.alias_generators import to_snake
from typing import Any, Optional, Union
import dotenv
import json
import logging
import os
import requests
import uuid

from .types import (
    PipelineRunError,
    PipelineRunResponse,
    NodeInput,
    PipelineRunRequest,
)


class APIError(Exception):
    def __init__(self, status: Union[int, str], message: str, details: Any = None):
        self.message = message
        self.status = status
        self.details = details

    def __str__(self):
        msg = "{0} ({1}): {2}"
        return msg.format(self.message, self.status, self.details)


class Laminar:
    _base_url = "https://api.lmnr.ai"

    def __init__(self, project_api_key: Optional[str] = None):
        self.project_api_key = project_api_key or os.environ.get("LMNR_PROJECT_API_KEY")
        if not self.project_api_key:
            dotenv_path = dotenv.find_dotenv(usecwd=True)
            self.project_api_key = dotenv.get_key(
                dotenv_path=dotenv_path, key_to_get="LMNR_PROJECT_API_KEY"
            )
        if not self.project_api_key:
            raise ValueError(
                "Please initialize the Laminar object with your project API key or set "
                "the LMNR_PROJECT_API_KEY environment variable in your environment or .env file"
            )

    def run(
        self,
        pipeline: str,
        inputs: dict[str, NodeInput],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {},
        parent_span_id: Optional[uuid.UUID] = None,
        trace_id: Optional[uuid.UUID] = None,
    ) -> PipelineRunResponse:
        """Runs the pipeline with the given inputs

        Args:
            pipeline (str): name of the Laminar pipeline
            inputs (dict[str, NodeInput]):
                inputs to the endpoint's target pipeline.
                Keys in the dictionary must match input node names
            env (dict[str, str], optional):
                Environment variables for the pipeline execution.
                Defaults to {}.
            metadata (dict[str, str], optional):
                any custom metadata to be stored
                with execution trace. Defaults to {}.
            parent_span_id (Optional[uuid.UUID], optional):
                parent span id for the resulting span.
                Must usually be SpanContext.id()
                Defaults to None.
            trace_id (Optional[uuid.UUID], optional):
                trace id for the resulting trace.
                Must usually be TraceContext.id()
                Defaults to None.

        Returns:
            PipelineRunResponse: response object containing the outputs

        Raises:
            ValueError: if project API key is not set
            PipelineRunError: if the endpoint run fails
        """
        if self.project_api_key is None:
            raise ValueError(
                "Please initialize the Laminar object with your project API key or set "
                "the LMNR_PROJECT_API_KEY environment variable"
            )
        try:
            request = PipelineRunRequest(
                inputs=inputs,
                pipeline=pipeline,
                env=env,
                metadata=metadata,
                parent_span_id=parent_span_id,
                trace_id=trace_id,
            )
        except Exception as e:
            raise ValueError(f"Invalid request: {e}")

        response = requests.post(
            self._base_url + "/v1/pipeline/run",
            data=json.dumps(request.to_dict()),
            headers=self._headers(),
        )
        if response.status_code != 200:
            raise PipelineRunError(response)
        try:
            resp_json = response.json()
            keys = list(resp_json.keys())
            for key in keys:
                value = resp_json[key]
                del resp_json[key]
                resp_json[to_snake(key)] = value
            return PipelineRunResponse(**resp_json)
        except Exception:
            raise PipelineRunError(response)

    def batch_post_traces(self, batch: list[Union[Span, Trace]]):
        log = logging.getLogger("laminar.client")
        url = self._base_url + "/v1/traces"
        data = json.dumps({"traces": [item.to_dict() for item in batch]})
        log.debug(f"making request to {url}")
        headers = self._headers()
        res = requests.post(url, data=data, headers=headers)

        if res.status_code == 200:
            log.debug("data uploaded successfully")

        return self._process_response(
            res, success_message="data uploaded successfully", return_json=False
        )

    def _process_response(
        self, res: requests.Response, success_message: str, return_json: bool = True
    ) -> Union[requests.Response, Any]:
        log = logging.getLogger("laminar.client")
        log.debug("received response: %s", res.text)
        if res.status_code in (200, 201):
            log.debug(success_message)
            if return_json:
                try:
                    return res.json()
                except json.JSONDecodeError:
                    log.error("Response is not valid JSON.")
                    raise APIError(res.status_code, "Invalid JSON response received")
            else:
                return res
        try:
            payload = res.json()
            log.error("received error response: %s", payload)
            raise APIError(res.status_code, payload)
        except (KeyError, ValueError):
            raise APIError(res.status_code, res.text)

    def _headers(self):
        return {
            "Authorization": "Bearer " + self.project_api_key,
            "Content-Type": "application/json",
        }



from .collector import ThreadManager
from .client import Laminar
from .providers import Provider
from .providers.fallback import FallbackProvider
from .tracing_types import EvaluateEvent, Event, Span, Trace
from .types import PipelineRunResponse
from .utils import PROVIDER_NAME_TO_OBJECT, is_async_iterator, is_iterator

from contextvars import ContextVar
from typing import Any, AsyncGenerator, Generator, Literal, Optional, Union
import atexit
import datetime
import dotenv
import inspect
import logging
import os
import pydantic
import uuid


_lmnr_stack_context: ContextVar[list[Union[Span, Trace]]] = ContextVar(
    "lmnr_stack_context", default=[]
)
_root_trace_id_context: ContextVar[Optional[str]] = ContextVar(
    "root_trace_id_context", default=None
)


class LaminarContextManager:
    _log = logging.getLogger("laminar.context_manager")

    def __init__(
        self,
        project_api_key: str = None,
        threads: int = 1,
        max_task_queue_size: int = 1000,
    ):
        self.project_api_key = project_api_key or os.environ.get("LMNR_PROJECT_API_KEY")
        if not self.project_api_key:
            dotenv_path = dotenv.find_dotenv(usecwd=True)
            self.project_api_key = dotenv.get_key(
                dotenv_path=dotenv_path, key_to_get="LMNR_PROJECT_API_KEY"
            )
        self.laminar = Laminar(project_api_key=self.project_api_key)
        self.thread_manager = ThreadManager(
            client=self.laminar,
            max_task_queue_size=max_task_queue_size,
            threads=threads,
        )
        # atexit executes functions last in first out, so we want to make sure
        # that we finalize the trace before thread manager is closed, so the updated
        # trace is sent to the server
        atexit.register(self._force_finalize_trace)

    def observe_start(
        self,
        # span attributes
        name: str,
        input: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        span_type: Literal["DEFAULT", "LLM"] = "DEFAULT",
        check_event_names: list[str] = None,
        # trace attributes
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
    ) -> Span:
        trace_id = _root_trace_id_context.get()
        if not trace_id:
            session_id = session_id or str(uuid.uuid4())
            trace_id = uuid.uuid4()
            trace = self.update_trace(
                id=trace_id,
                user_id=user_id,
                session_id=session_id,
                release=release,
            )
            _root_trace_id_context.set(trace.id)
            _lmnr_stack_context.set([trace])

        parent = _lmnr_stack_context.get()[-1] if _lmnr_stack_context.get() else None
        parent_span_id = parent.id if isinstance(parent, Span) else None
        span = self.create_span(
            name=name,
            trace_id=trace_id,
            input=input,
            metadata=metadata,
            attributes=attributes,
            parent_span_id=parent_span_id,
            span_type=span_type,
            check_event_names=check_event_names,
        )
        stack = _lmnr_stack_context.get()
        _lmnr_stack_context.set(stack + [span])
        return span

    def observe_end(
        self,
        span: Span,
        provider_name: str = None,
        result: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        error: Optional[Exception] = None,
    ) -> Any:
        stack = _lmnr_stack_context.get()
        if not stack:
            return
        provider = PROVIDER_NAME_TO_OBJECT.get(provider_name, FallbackProvider())
        new_stack = stack[:-1]
        _lmnr_stack_context.set(new_stack)

        if len(new_stack) == 1 and isinstance(stack[0], Trace):
            trace = stack[0]
            self.update_trace(
                id=trace.id,
                user_id=trace.userId,
                session_id=trace.sessionId,
                release=trace.release,
                metadata=metadata,
            )
            _root_trace_id_context.set(None)
            _lmnr_stack_context.set([])

        if error is not None:
            self.update_current_trace(success=False)

        if inspect.isgenerator(result) or is_iterator(result):
            return self._collect_generator_result(
                provider=provider,
                generator=result,
                span=span,
                metadata=metadata,
                attributes=attributes,
            )
        elif inspect.isasyncgen(result) or is_async_iterator(result):
            return self._collect_async_generator_result(
                provider=provider,
                generator=result,
                span=span,
                metadata=metadata,
                attributes=attributes,
            )
        if span.spanType == "LLM" and error is None:
            attributes = self._extract_llm_attributes_from_response(
                provider=provider, response=result
            )
        return self._finalize_span(
            span,
            provider=provider,
            result=error or result,
            metadata=metadata,
            attributes=attributes,
        )

    def update_current_span(
        self,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        evaluate_events: list[EvaluateEvent] = None,
        override: bool = False,
    ):
        stack = _lmnr_stack_context.get()
        if not stack:
            return
        span = stack[-1]
        new_metadata = (
            metadata if override else {**(span.metadata or {}), **(metadata or {})}
        )
        new_evaluate_events = (
            evaluate_events
            if override
            else span.evaluateEvents + (evaluate_events or [])
        )
        new_attributes = (
            attributes
            if override
            else {**(span.attributes or {}), **(attributes or {})}
        )
        self.update_span(
            span=span,
            metadata=new_metadata,
            evaluate_events=new_evaluate_events,
            attributes=new_attributes,
        )

    def update_current_trace(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        success: bool = True,
    ):
        existing_trace = (
            _lmnr_stack_context.get()[0] if _lmnr_stack_context.get() else None
        )
        if not existing_trace:
            return
        self.update_trace(
            id=existing_trace.id,
            user_id=user_id or existing_trace.userId,
            session_id=session_id or existing_trace.sessionId,
            release=release or existing_trace.release,
            metadata=metadata or existing_trace.metadata,
            success=success if success is not None else existing_trace.success,
        )

    def update_trace(
        self,
        id: uuid.UUID,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        success: bool = True,
    ) -> Trace:
        trace = Trace(
            id=id,
            user_id=user_id,
            session_id=session_id,
            release=release,
            metadata=metadata,
            success=success,
        )
        self._add_observation(trace)
        return trace

    def create_span(
        self,
        name: str,
        trace_id: uuid.UUID,
        start_time: Optional[datetime.datetime] = None,
        span_type: Literal["DEFAULT", "LLM"] = "DEFAULT",
        id: Optional[uuid.UUID] = None,
        parent_span_id: Optional[uuid.UUID] = None,
        input: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        check_event_names: list[str] = None,
    ) -> Span:
        """Internal method to create a span object. Use `ObservationContext.span` instead."""
        span = Span(
            name=name,
            trace_id=trace_id,
            start_time=start_time or datetime.datetime.now(datetime.timezone.utc),
            id=id,
            parent_span_id=parent_span_id,
            input=input,
            metadata=metadata,
            attributes=attributes,
            span_type=span_type,
            evaluate_events=check_event_names or [],
        )
        return span

    def update_span(
        self,
        span: Span,
        finalize: bool = False,
        input: Optional[Any] = None,
        end_time: Optional[datetime.datetime] = None,
        output: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        evaluate_events: Optional[list[EvaluateEvent]] = None,
        override: bool = False,
    ) -> Span:
        """Internal method to update a span object. Use `SpanContext.update()` instead."""
        span.update(
            input=input or span.input,
            output=output or span.output,
            end_time=end_time,
            metadata=metadata,
            attributes=attributes,
            evaluate_events=evaluate_events,
            override=override,
        )
        if finalize:
            self._add_observation(span)
        return span

    def event(
        self,
        name: str,
        value: Optional[Union[str, int, float, bool]] = None,
        timestamp: Optional[datetime.datetime] = None,
    ):
        span = _lmnr_stack_context.get()[-1] if _lmnr_stack_context.get() else None
        if not span or not isinstance(span, Span):
            self._log.warning(f"No active span to send event. Ignoring event. {name}")
            return
        event = Event(
            name=name,
            span_id=span.id,
            timestamp=timestamp,
            value=value,
        )
        span.add_event(event)

    def evaluate_event(self, name: str, data: str):
        stack = _lmnr_stack_context.get()
        if not stack or not isinstance(stack[-1], Span):
            self._log.warning(
                f"No active span to add check event. Ignoring event. {name}"
            )
            return
        stack[-1].evaluateEvents.append(
            EvaluateEvent(
                name=name,
                data=data,
                timestamp=datetime.datetime.now(datetime.timezone.utc),
            )
        )

    def run_pipeline(
        self,
        pipeline: str,
        inputs: dict[str, Any],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {},
    ) -> PipelineRunResponse:
        span = _lmnr_stack_context.get()[-1] if _lmnr_stack_context.get() else None
        span_id = span.id if isinstance(span, Span) else None
        trace = _lmnr_stack_context.get()[0] if _lmnr_stack_context.get() else None
        trace_id = trace.id if isinstance(trace, Trace) else None
        return self.laminar.run(
            pipeline=pipeline,
            inputs=inputs,
            env=env,
            metadata=metadata,
            parent_span_id=span_id,
            trace_id=trace_id,
        )

    def _force_finalize_trace(self):
        # TODO: flush in progress spans as error?
        pass

    def _add_observation(self, observation: Union[Span, Trace]) -> bool:
        return self.thread_manager.add_task(observation)

    def _extract_llm_attributes_from_response(
        self,
        provider: Provider,
        response: Union[str, dict[str, Any], pydantic.BaseModel],
    ) -> dict[str, Any]:
        return provider.extract_llm_attributes_from_response(response)

    def _stream_list_to_dict(
        self, provider: Provider, response: list[Any]
    ) -> dict[str, Any]:
        return provider.stream_list_to_dict(response)

    def _extract_llm_output(
        self,
        provider: Provider,
        result: Union[dict[str, Any], pydantic.BaseModel],
    ) -> str:
        return provider.extract_llm_output(result)

    def _finalize_span(
        self,
        span: Span,
        provider: Provider = None,
        result: Optional[Any] = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
    ) -> Any:
        self.update_span(
            span=span,
            finalize=True,
            output=(
                result
                if span.spanType != "LLM"
                else self._extract_llm_output(provider, result)
            ),
            metadata=metadata,
            attributes=attributes,
        )
        return result

    def _collect_generator_result(
        self,
        generator: Generator,
        span: Span,
        provider: Provider = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
    ) -> Generator:
        items = []
        try:
            for item in generator:
                items.append(item)
                yield item

        finally:
            output = items
            if all(isinstance(item, str) for item in items):
                output = "".join(items)
            if span.spanType == "LLM":
                collected = self._stream_list_to_dict(
                    provider=provider, response=output
                )
                attributes = self._extract_llm_attributes_from_response(
                    provider=provider, response=collected
                )
            self._finalize_span(
                span=span,
                provider=provider,
                result=collected,
                metadata=metadata,
                attributes=attributes,
            )

    async def _collect_async_generator_result(
        self,
        generator: AsyncGenerator,
        span: Span,
        provider: Provider = None,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
    ) -> AsyncGenerator:
        items = []
        try:
            async for item in generator:
                items.append(item)
                yield item

        finally:
            output = items
            if all(isinstance(item, str) for item in items):
                output = "".join(items)
            if span.spanType == "LLM":
                collected = self._stream_list_to_dict(
                    provider=provider, response=output
                )
                attributes = self._extract_llm_attributes_from_response(
                    provider=provider, response=collected
                )
            self._finalize_span(
                span=span,
                provider=provider,
                result=collected,
                metadata=metadata,
                attributes=attributes,
            )


# TODO: add lock for thread safety
class LaminarSingleton:
    _instance = None
    _l: Optional[LaminarContextManager] = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(LaminarSingleton, cls).__new__(cls)
        return cls._instance

    def get(cls, *args, **kwargs) -> LaminarContextManager:
        if not cls._l:
            cls._l = LaminarContextManager(*args, **kwargs)
        return cls._l



import logging
import json
import pydantic
import typing

logger = logging.getLogger("lmnr.sdk.tracing.providers.utils")


def parse_or_dump_to_dict(
    obj: typing.Union[pydantic.BaseModel, dict[str, typing.Any], str]
) -> dict[str, typing.Any]:
    if isinstance(obj, pydantic.BaseModel):
        return obj.model_dump()
    if isinstance(obj, str):
        try:
            parsed = json.loads(obj)
            if isinstance(parsed, dict):
                return parsed
            else:
                logging.warning(
                    f"Expected a dict, but got: {type(parsed)}. Returning empty dict."
                )
                return {}
        except Exception as e:
            logging.error(f"Error parsing string: {e}")
            return {}

    if isinstance(obj, dict):
        return obj
    logging.warning(
        f"Expected a dict, BaseModel, or str, but got {type(obj)}. Returning empty dict."
    )
    return {}



import datetime
import functools
from typing import Any, Callable, Literal, Optional, Union


from .context import LaminarSingleton
from .providers.fallback import FallbackProvider
from ..semantic_conventions.gen_ai_spans import PROVIDER
from .types import NodeInput, PipelineRunResponse
from .utils import (
    PROVIDER_NAME_TO_OBJECT,
    get_input_from_func_args,
    is_async,
    is_method,
)


class LaminarDecorator:
    def observe(
        self,
        *,
        name: Optional[str] = None,
        span_type: Optional[Literal["DEFAULT", "LLM"]] = "DEFAULT",
        capture_input: bool = True,
        capture_output: bool = True,
        release: Optional[str] = None,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
    ):
        """The main decorator entrypoint for Laminar. This is used to wrap functions and methods to create spans.

        Args:
            name (Optional[str], optional): Name of the span. Function name is used if not specified. Defaults to None.
            span_type (Literal[&quot;DEFAULT&quot;, &quot;LLM&quot;], optional): Type of this span. Prefer `wrap_llm_call` instead of specifying
                                                                                 this as "LLM" . Defaults to "DEFAULT".
            capture_input (bool, optional): Whether to capture input parameters to the function. Defaults to True.
            capture_output (bool, optional): Whether to capture returned type from the function. Defaults to True.
            release (Optional[str], optional): Release version of your app. Useful for further grouping and analytics. Defaults to None.
            user_id (Optional[str], optional): Custom user_id of your user. Useful for grouping and further analytics. Defaults to None.
            session_id (Optional[str], optional): Custom session_id for your session. Random UUID is generated on Laminar side, if not specified.
                                                  Defaults to None.

        Raises:
            Exception: re-raises the exception if the wrapped function raises an exception

        Returns:
            Any: Returns the result of the wrapped function
        """
        context_manager = LaminarSingleton().get()

        def decorator(func: Callable):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                span = context_manager.observe_start(
                    name=name or func.__name__,
                    span_type=span_type,
                    input=(
                        get_input_from_func_args(func, is_method(func), args, kwargs)
                        if capture_input
                        else None
                    ),
                    user_id=user_id,
                    session_id=session_id,
                    release=release,
                )
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    context_manager.observe_end(result=None, span=span, error=e)
                    raise e
                context_manager.observe_end(
                    result=result if capture_output else None, span=span
                )
                return result

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                span = context_manager.observe_start(
                    name=name or func.__name__,
                    span_type=span_type,
                    input=(
                        get_input_from_func_args(func, is_method(func), args, kwargs)
                        if capture_input
                        else None
                    ),
                    user_id=user_id,
                    session_id=session_id,
                    release=release,
                )
                try:
                    result = await func(*args, **kwargs)
                except Exception as e:
                    context_manager.observe_end(result=None, span=span, error=e)
                    raise e
                context_manager.observe_end(
                    result=result if capture_output else None, span=span
                )
                return result

            return async_wrapper if is_async(func) else wrapper

        return decorator

    def update_current_span(
        self,
        metadata: Optional[dict[str, Any]] = None,
        attributes: Optional[dict[str, Any]] = None,
        override: bool = False,
    ):
        """Update the current span with any optional metadata.

        Args:
            metadata (Optional[dict[str, Any]], optional): metadata to the span. Defaults to None.
            override (bool, optional): Whether to override the existing metadata. If False, metadata is merged with the existing metadata. Defaults to False.
        """
        laminar = LaminarSingleton().get()
        laminar.update_current_span(
            metadata=metadata, attributes=attributes, override=override
        )

    def update_current_trace(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        release: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ):
        """Update the current trace with any optional metadata.

        Args:
            user_id (Optional[str], optional): Custom user_id of your user. Useful for grouping and further analytics. Defaults to None.
            session_id (Optional[str], optional): Custom session_id for your session. Random UUID is generated on Laminar side, if not specified.
                                                  Defaults to None.
            release (Optional[str], optional): Release version of your app. Useful for further grouping and analytics. Defaults to None.
            metadata (Optional[dict[str, Any]], optional): metadata to the trace. Defaults to None.
        """
        laminar = LaminarSingleton().get()
        laminar.update_current_trace(
            user_id=user_id, session_id=session_id, release=release, metadata=metadata
        )

    def event(
        self,
        name: str,
        value: Optional[Union[str, int, float, bool]] = None,
        timestamp: Optional[datetime.datetime] = None,
    ):
        """Associate an event with the current span

        Args:
            name (str): name of the event. Must be predefined in the Laminar events page.
            value (Optional[Union[str, int, float, bool]], optional): value of the event. Must match range definition in Laminar events page. Defaults to None.
            timestamp (Optional[datetime.datetime], optional): If you need custom timestamp. If not specified, current time is used. Defaults to None.
        """
        laminar = LaminarSingleton().get()
        laminar.event(name, value=value, timestamp=timestamp)

    def evaluate_event(self, name: str, data: str):
        """Evaluate an event with the given name and data. The event value will be assessed by the Laminar evaluation engine.
        Data is passed as an input to the agent, so you need to specify which data you want to evaluate. Most of the times,
        this is an output of the LLM generation, but sometimes, you may want to evaluate the input or both. In the latter case,
        concatenate the input and output annotating with natural language.

        Args:
            name (str): Name of the event. Must be predefined in the Laminar events page.
            data (str): Data to be evaluated. Typically the output of the LLM generation.
        """
        laminar = LaminarSingleton().get()
        laminar.evaluate_event(name, data)

    def run(
        self,
        pipeline: str,
        inputs: dict[str, NodeInput],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {},
    ) -> PipelineRunResponse:
        """Run the laminar pipeline with the given inputs. Pipeline must be defined in the Laminar UI and have a target version.

        Args:
            pipeline (str): pipeline name
            inputs (dict[str, NodeInput]): Map from input node name to input value
            env (dict[str, str], optional): Environment variables for the pipeline executions. Typically contains API keys. Defaults to None.
            metadata (dict[str, str], optional): Any additional data to associate with the resulting span. Defaults to None.

        Returns:
            PipelineRunResponse: Response from the pipeline execution
        """
        laminar = LaminarSingleton().get()
        return laminar.run_pipeline(pipeline, inputs, env, metadata)


def wrap_llm_call(func: Callable, name: str = None, provider: str = None) -> Callable:
    """Wrap an LLM call with Laminar observability. This is a convenience function that does the same as `@observe()`, plus
    a few utilities around LLM-specific things, such as counting tokens and recording model params.

    Example usage:
    ```python
    wrap_llm_call(client.chat.completions.create)(
        model="gpt-4o-mini",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Hello"},
        ],
        stream=True,
    )
    ```

    Args:
        func (Callable): The function to wrap
        name (str, optional): Name of the resulting span. Default "{provider name} completion" if not specified. Defaults to None.
        provider (str, optional): LLM model provider, e.g. openai, anthropic. This is needed to help us correctly parse
                                  things like token usage. If not specified, we infer it from the name of the package,
                                  where the function is imported from. Defaults to None.

    Raises:
        Exctption: re-raises the exception if the wrapped function raises an exception

    Returns:
        Callable: the wrapped function
    """
    laminar = LaminarSingleton().get()
    # Simple heuristic to determine the package from where the LLM call is imported.
    # This works for major providers, but will likely make no sense for custom providers.
    provider_name = (
        provider.lower().strip() if provider else func.__module__.split(".")[0]
    )
    provider_module = PROVIDER_NAME_TO_OBJECT.get(provider_name, FallbackProvider())
    name = name or f"{provider_module.display_name()} completion"

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        inp = kwargs.get("messages")
        attributes = (
            provider_module.extract_llm_attributes_from_args(args, kwargs)
            if provider_module
            else {}
        )
        attributes[PROVIDER] = provider_name
        span = laminar.observe_start(
            name=name, span_type="LLM", input=inp, attributes=attributes
        )
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            laminar.observe_end(
                result=None, span=span, error=e, provider_name=provider_name
            )
            raise e
        return laminar.observe_end(
            result=result, span=span, provider_name=provider_name
        )

    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        inp = kwargs.get("messages")
        attributes = (
            provider_module.extract_llm_attributes_from_args(args, kwargs)
            if provider_module
            else {}
        )
        attributes[PROVIDER] = provider_name
        span = laminar.observe_start(
            name=name, span_type="LLM", input=inp, attributes=attributes
        )
        try:
            result = await func(*args, **kwargs)
        except Exception as e:
            laminar.observe_end(
                result=None, span=span, error=e, provider_name=provider_name
            )
            raise e
        return laminar.observe_end(
            result=result, span=span, provider_name=provider_name
        )

    return async_wrapper if is_async(func) else wrapper


lmnr_context = LaminarDecorator()
observe = lmnr_context.observe



# source: https://github.com/open-telemetry/semantic-conventions/blob/main/docs/gen-ai/gen-ai-spans.md
# last updated: 2024-08-26

REQUEST_MODEL: str = "gen_ai.request.model"
RESPONSE_MODEL: str = "gen_ai.response.model"
PROVIDER: str = "gen_ai.system"
INPUT_TOKEN_COUNT: str = "gen_ai.usage.input_tokens"
OUTPUT_TOKEN_COUNT: str = "gen_ai.usage.output_tokens"
TOTAL_TOKEN_COUNT: str = "gen_ai.usage.total_tokens"  # custom, not in the spec
# https://github.com/openlit/openlit/blob/main/sdk/python/src/openlit/semcov/__init__.py
COST: str = "gen_ai.usage.cost"

OPERATION: str = "gen_ai.operation.name"

FREQUENCY_PENALTY: str = "gen_ai.request.frequency_penalty"
TEMPERATURE: str = "gen_ai.request.temperature"
MAX_TOKENS: str = "gen_ai.request.max_tokens"
PRESENCE_PENALTY: str = "gen_ai.request.presence_penalty"
STOP_SEQUENCES: str = "gen_ai.request.stop_sequences"
TEMPERATURE: str = "gen_ai.request.temperature"
TOP_P: str = "gen_ai.request.top_p"
TOP_K: str = "gen_ai.request.top_k"

# https://github.com/openlit/openlit/blob/main/sdk/python/src/openlit/semcov/__init__.py
STREAM: str = "gen_ai.request.is_stream"

FINISH_REASONS = "gen_ai.response.finish_reasons"

__all__ = [
    "REQUEST_MODEL",
    "RESPONSE_MODEL",
    "PROVIDER",
    "INPUT_TOKEN_COUNT",
    "OUTPUT_TOKEN_COUNT",
    "TOTAL_TOKEN_COUNT",
    "COST",
    "OPERATION",
    "FREQUENCY_PENALTY",
    "TEMPERATURE",
    "MAX_TOKENS",
    "PRESENCE_PENALTY",
    "STOP_SEQUENCES",
    "TEMPERATURE",
    "TOP_P",
    "TOP_K",
    "STREAM",
    "FINISH_REASONS",
]



from .client import APIError, Laminar
from .tracing_types import Span, Trace

from queue import Queue, Empty, Full
from typing import Union

import atexit
import backoff
import logging
import time
import threading


class Collector(threading.Thread):
    _log = logging.getLogger("laminar.collector")
    _queue: Queue[Union[Span, Trace]]
    _client: Laminar
    _flush_interval: float

    def __init__(
        self,
        queue: Queue[Union[Span, Trace]],
        client: Laminar,
        flush_interval: float = 5.0,
    ):
        super().__init__()
        self.daemon = True
        self._queue = queue
        self.running = True
        self._flush_interval = flush_interval
        self._client = client

    def run(self):
        """Runs the collector."""
        self._log.debug("collector is running...")
        while self.running:
            self.upload()

    def upload(self):
        """Upload the next batch of items, return whether successful."""
        batch = self._next()
        if len(batch) == 0:
            return

        try:
            self._upload_batch(batch)
        except Exception as e:
            self._log.exception("error uploading: %s", e)
        finally:
            # mark items as acknowledged from queue
            for _ in batch:
                self._queue.task_done()

    def pause(self):
        self.running = False

    def _next(self):
        items = []
        start_time = time.monotonic()

        while True:
            elapsed = time.monotonic() - start_time
            if elapsed >= self._flush_interval:
                break
            try:
                item = self._queue.get(
                    block=True, timeout=self._flush_interval - elapsed
                )
                items.append(item)
            except Empty:
                break
        return items

    def _upload_batch(self, batch: list[Union[Trace, Span]]):
        self._log.debug("uploading batch of %d items", len(batch))

        @backoff.on_exception(backoff.expo, Exception, max_tries=5)
        def execute_task_with_backoff(batch: list[Union[Trace, Span]]):
            try:
                self._client.batch_post_traces(batch=batch)
            except Exception as e:
                if (
                    isinstance(e, APIError)
                    and 400 <= int(e.status) < 500
                    and int(e.status) != 429
                ):
                    self._log.warn(
                        f"Received {e.status} error by Laminar server, not retrying: {e.message}"
                    )
                    return

                raise e

        execute_task_with_backoff(batch)
        self._log.debug("successfully uploaded batch of %d items", len(batch))


class ThreadManager:
    _log = logging.getLogger("laminar.task_manager")
    _queue: Queue[Union[Span, Trace]]
    _client: Laminar
    _max_task_queue_size: int
    _flush_interval: float
    _collectors: list[Collector] = []
    _threads: int

    def __init__(
        self,
        client: Laminar,
        flush_interval: float = 2.0,
        max_task_queue_size: int = 1000,
        threads: int = 1,
    ):
        self._max_task_queue_size = max_task_queue_size
        self._threads = threads
        self._queue = Queue(maxsize=self._max_task_queue_size)
        self._flush_interval = flush_interval
        self._client = client
        for _ in range(self._threads):
            collector = Collector(
                queue=self._queue,
                client=self._client,
                flush_interval=flush_interval,
            )
            self._collectors.append(collector)
            collector.start()
        atexit.register(self.join)

    def add_task(self, event: Union[Span, Trace]) -> bool:
        try:
            self._queue.put(event, block=False)
            return True
        except Full:
            self._log.warning("queue is full")
            return False
        except Exception as e:
            self._log.exception(f"Exception in adding task {e}")

            return False

    def flush(self):
        """Forces a flush from the internal queue to the server"""
        self._log.debug("flushing queue")
        queue = self._queue
        size = queue.qsize()
        queue.join()
        # Note that this message may not be precise, because of threading.
        self._log.debug("successfully flushed about %s items.", size)

    def join(self):
        """Ends the collector threads once the queue is empty.
        Blocks execution until finished
        """
        self._log.debug(f"joining {len(self._collectors)} collector threads")

        # pause all collectors before joining them so we don't have to wait for multiple
        # flush intervals to join them all.
        for collector in self._collectors:
            collector.pause()

        for i, collector in enumerate(self._collectors):
            try:
                collector.join()
            except RuntimeError:
                # collector thread has not started
                pass

            self._log.debug(f"collector thread {i} joined")

    def shutdown(self):
        """Flush all messages and cleanly shutdown the client"""
        self._log.debug("shutdown initiated")

        self.flush()
        self.join()

        self._log.debug("shutdown completed")
